/*
    This file is part of random-image.

    random-image is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    random-image is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with random-image.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "MainWindow.h"
#include <QFileDialog>
#include <QApplication>
#include <QListWidget>
#include <QListWidgetItem>
#include <QIcon>
#include <QPixmap>
#include <QSize>
#include <QFile>
#include <QDir>
#include <time.h>
#include <QMessageBox>
#include <iostream>

MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent)
{
    qsrand(time(NULL));
    setupUi(this);
    arreter->hide();
    m_image.empty();
    curImage(&m_image);
    m_progressBar = new QProgressBar;
    statusbar->addWidget(m_progressBar);
    aboutWindow = new About(this);
}

void MainWindow::curImage(Network* net)
{
    current_image->setPixmap(QPixmap::fromImage(net->image(current_image->width(), current_image->height())));
}

void MainWindow::on_raz_clicked()
{
    m_image.empty();
    curImage(&m_image);
}

void MainWindow::on_miniatures_itemDoubleClicked(QListWidgetItem*)
{
    curImage(&m_network_miniatures[miniatures->currentRow()]);
    m_image = m_network_miniatures[miniatures->currentRow()];
}

void MainWindow::on_demarrer_clicked()
{
    miniatures->clear();
    m_network_miniatures.clear();
    for(unsigned int i = 0; i < m_miniatures_threads.count(); i++)
        delete m_miniatures_threads[i];
    m_miniatures_threads.clear();
    miniatures->setIconSize(QSize(miniature_width->value(), miniature_height->value()));
    m_progressBar->setRange(0, nb_images->value());
    if(auto_modifications->isChecked())
    {
        unsigned int value = qrand() % 50 + 30 - sqrt(m_image.getComplexity()) * 16;
        modifications->setValue(value);
    }

    for(unsigned int i = 0; i < QThread::idealThreadCount(); i++)
    {
        MiniatureThread* t = new MiniatureThread(this, i);
        t->start();
        m_miniatures_threads.append(t);
    }
    arreter->show();
    demarrer->hide();
}

void MainWindow::receive_miniature(Network net, QImage image)
{
    miniatures->addItem(new QListWidgetItem(QIcon(QPixmap::fromImage(image)), QString::number(net.getComplexity())));
    m_network_miniatures.append(net);
    m_progressBar->setValue(m_network_miniatures.count());
    if(nb_images->value() <= m_network_miniatures.count())
    {
        arreter->hide();
        demarrer->show();
        m_progressBar->reset();
    }
}


void MainWindow::on_arreter_clicked()
{
    for(unsigned int i = 0; i < m_miniatures_threads.count(); i++)
        m_miniatures_threads[i]->pleaseStop();
    arreter->hide();
    demarrer->show();
}

void MainWindow::on_generer_image_clicked()
{
    qint16 width = QApplication::desktop()->width();
    qint16 height = QApplication::desktop()->height();
    QString fname;
    qint32 i = 0;
    do
    {
        i++;
        fname = QDir::currentPath() + "/" + QString::number(i) + ".png";
    }
    while(QFile::exists(fname));
    QString fichier = QFileDialog::getSaveFileName(this, "Enregistrer une image", fname, "Image (*.png *.jpg *.jpeg *.bmp);;Tous les fichiers (*.*)");
    if(fichier == "")
        return;
    QImage img = m_image.image(width, height, true, this);
    if(img != QImage()) // si on a annulé, ça renvoie un QImage()
        img.save(fichier, 0, 90);
}

void MainWindow::on_ouvrir_clicked()
{
    QString fichier = QFileDialog::getOpenFileName(this, "Ouvrir la source d'une image", QString(), "Source d'image (*.ris);;Tous les fichiers (*.*)");
    if(fichier == "")
        return;
    if(! m_image.readFromFile(fichier))
        QMessageBox::warning(this, "Erreur", "Le fichier est invalide ou corrompu !");
    curImage(&m_image);
}

void MainWindow::on_sauver_clicked()
{
    QString fichier = QFileDialog::getSaveFileName(this, "Enregistrer la source d'une image", QString(), "Source d'image (*.ris);;Tous les fichiers (*.*)");
    if(fichier == "")
        return;
    m_image.saveToFile(fichier);
}

void MainWindow::on_autogen_clicked()
{
    // fonction qui doit prendre une image, et essayer d'obtenir un RdN qui correspond à peu près à l'image
    // Problème : ça ne donne AUCUN résultat
    // en plus le code est dégueueuuuuuuu
    QString fichier = QFileDialog::getOpenFileName(this, "Ouvrir une image", QString(), "Images (*.png *.jpg *.gif);;Tous les fichiers (*.*)");
    if(fichier == "")
        return;
    QImage original_image(fichier);
    original_image = original_image.scaled(QSize(getImageWidth(), getImageHeight()));
    miniatures->clear();
    m_image.empty();
    unsigned long actual_dist = 99999999;
    while(true)
    {
        Network tmp_net = m_image;
        tmp_net.mutate(getMutations(), getModifications());
        QImage tmp_image = tmp_net.image(getImageWidth(), getImageHeight());

        unsigned long distance = 0;
        for(unsigned int x = 0; x < getImageWidth(); x++)
        {
            QApplication::processEvents();
            for(unsigned int y = 0; y < getImageHeight(); y++)
            {
                QRgb couleur1 = original_image.pixel(x, y);
                QRgb couleur2 = tmp_image.pixel(x, y);
                distance += abs(qRed(couleur1) - qRed(couleur2));
                distance += abs(qGreen(couleur1) - qGreen(couleur2));
                distance += abs(qBlue(couleur1) - qBlue(couleur2));
                //distance += abs(qGray(couleur1) - qGray(couleur2));
            }
        }

        std::cout << actual_dist << " " << distance << std::endl;
        if(actual_dist > distance)
        {
            m_image = tmp_net;
            curImage(&m_image);
            miniatures->addItem(new QListWidgetItem(QIcon(QPixmap::fromImage(tmp_image)), QString::number(tmp_net.getComplexity())));
            m_network_miniatures.append(tmp_net);
            actual_dist = distance;
        }
    }
}

void MiniatureThread::run()
{
    qsrand(time(NULL) + begin);
    qRegisterMetaType<Network>("Network");
    QObject::connect(this, SIGNAL(send_miniature(Network, QImage)), mainWindow, SLOT(receive_miniature(Network, QImage)));

    unsigned int iThreadCount = QThread::idealThreadCount();
    for(unsigned int i = begin; i < mainWindow->getNbImages(); i += iThreadCount)
    {
        if(stop)
            return;
        Network network = mainWindow->getCurrentNetwork();
        network.mutate(mainWindow->getMutations(), mainWindow->getModifications());
        emit send_miniature(network, network.image(mainWindow->getImageWidth(), mainWindow->getImageHeight()));
    }
}

qint16 MainWindow::getMutations()
{
    return mutations->value() + (qrand() % 4) - 2;
}
