/*
    This file is part of random-image.

    random-image is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    random-image is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with random-image.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HEADER_MAINWINDOW
#define HEADER_MAINWINDOW

#include <QtGui>
#include <QList>
#include <QThread>
#include <QProgressBar>
#include <QDialog>
#include "ui_mainwindow.h"
#include "ui_about.h"
#include "network.h"

class MainWindow;

class MiniatureThread : public QThread
{
    Q_OBJECT

    public:
        MiniatureThread(MainWindow* mainWindow, unsigned int begin) : mainWindow(mainWindow), begin(begin), stop(false) {};
        void pleaseStop() { stop = true; };

    signals:
        void send_miniature(Network net, QImage image);

    protected:
        void run();

    private:
        MainWindow* mainWindow;
        unsigned int begin;
        bool stop;
};

class About : public QDialog, private Ui::About
{
    Q_OBJECT

    public:
        About(QWidget* parent = 0) : QDialog(parent) { setupUi(this); };
};

class MainWindow : public QMainWindow, private Ui::MainWindow
{
    Q_OBJECT

    public:
        MainWindow(QWidget *parent = 0);
        qint16 getMutations();
        qint16 getNbImages()  { return nb_images->value(); };
        qint16 getImageHeight()  { return miniature_height->value(); };
        qint16 getImageWidth()  { return miniature_width->value(); };
        qint16 getModifications()  { return modifications->value(); };
        Network getCurrentNetwork() { return m_image; };
        void curImage(Network*);

    private:
        Network m_image;
        About* aboutWindow;
        QList<QLabel*> m_miniatures;
        QList<Network> m_network_miniatures;
        QList<MiniatureThread*> m_miniatures_threads;
        QProgressBar* m_progressBar;

    private slots:
        void on_demarrer_clicked();
        void on_generer_image_clicked();
        void on_ouvrir_clicked();
        void on_sauver_clicked();
        void on_arreter_clicked();
        void on_miniatures_itemDoubleClicked(QListWidgetItem*);
        void on_raz_clicked();
        void on_autogen_clicked();
        void receive_miniature(Network net, QImage image);
        void on_about_clicked() { aboutWindow->show(); };
};


#endif
