/*
    This file is part of random-image.

    random-image is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    random-image is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with random-image.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "network.h"
#include <math.h>
#include <QColor>
#include <QQueue>
#include <QProgressDialog>
#include <QApplication>
#include <QFile>
#include <iostream> //////////////////////////////////////////////////////
#define m_nbEntrees 5
#define SAFE_DIV(a, b) ((b == 0) ? 0 : a / b)

Network::Network()
{
    empty();
}

Network::~Network()
{
}

bool Network::readFromFile(QString fileName)
{
    QList<Neurone> net;

    QFile file(fileName);
    if(! file.open(QIODevice::ReadOnly))
        return false;

    qint16 nbr_neurones;
    file.read(reinterpret_cast<char*>(&nbr_neurones), sizeof(nbr_neurones));
    for(qint16 i = 0; i < nbr_neurones; i++)
    {
        Neurone n;
        file.read(reinterpret_cast<char*>(&n.type), sizeof(n.type));
        if(n.type > 5)
            return false;
        file.read(reinterpret_cast<char*>(&n.seuil), sizeof(n.seuil));
        n.valeur = 0;
        qint16 nbr_synapses;
        file.read(reinterpret_cast<char*>(&nbr_synapses), sizeof(nbr_synapses));
        if(nbr_synapses < 0)
            return false;
        n.synapses = std::vector<Synapse>();
        for(qint16 j = 0; j < nbr_synapses; j++)
        {
            Synapse s;
            file.read(reinterpret_cast<char*>(&s.neurone_id), sizeof(s.neurone_id));
            file.read(reinterpret_cast<char*>(&s.poids), sizeof(s.poids));
            if(s.neurone_id > nbr_neurones - 1)
                return false;
            n.synapses.push_back(s);
        }
        net.append(n);
    }

    if(! isValidNet(&net))
        return false;
    m_network = net;
    return true;
}

bool Network::saveToFile(QString fileName)
{
    QFile file(fileName);
    if(! file.open(QIODevice::WriteOnly | QIODevice::Truncate))
        return false;

    qint16 nbr_neurones = m_network.count();
    file.write(reinterpret_cast<char*>(&nbr_neurones), sizeof(nbr_neurones));
    for(qint16 i = 0; i < m_network.count(); i++)
    {
        qint16 nbr_synapses = m_network[i].synapses.size();
        file.write(reinterpret_cast<char*>(&m_network[i].type), sizeof(m_network[i].type));
        file.write(reinterpret_cast<char*>(&m_network[i].seuil), sizeof(m_network[i].seuil));
        file.write(reinterpret_cast<char*>(&nbr_synapses), sizeof(nbr_synapses));
        for(qint16 j = 0; j < m_network[i].synapses.size(); j++)
        {
            file.write(reinterpret_cast<char*>(&m_network[i].synapses[j].neurone_id), sizeof(m_network[i].synapses[j].neurone_id));
            file.write(reinterpret_cast<char*>(&m_network[i].synapses[j].poids), sizeof(m_network[i].synapses[j].poids));
        }
    }
    return true;
}

void Network::mutate(qint16 mutations, qint16 modifications)
{
    unsigned int nid;
    unsigned int nb_mutations = 0;
    while(nb_mutations < mutations)
    {
        double choix = 30 * sqrt(-2. * log((rand() % 1000) / 1000.)) * sin(2. * M_PI * (rand() % 1000) / 1000.) + modifications;
        // voilà, choix vaut un nombre, probablement centré sur modifications :)
        // il vaut -70 ~ 70 + modifications
        // PS : le sqrt donne entre -25 et 25


        if(choix < -35)
        {
            // déletion

            if(rand() % 2 == 0) // suppression d'un synapse à un neurone
            {
                nid = rand() % m_network.size();
                if(m_network[nid].synapses.size() < 2)
                    continue;
                m_network[nid].synapses.erase(m_network[nid].synapses.begin() + (rand() % m_network[nid].synapses.size()));
            }
            else
            {
                if(m_network.size() < m_nbEntrees + 2)
                    continue;
                nid = rand() % m_network.size();
                std::vector<Synapse> sorties = m_network[nid].synapses;
                m_network.erase(m_network.begin() + nid);
                // on recrée les connexions pour les neurones qui y étaient branchés et on décrémente les ID vers un neurone supérieur
                for(unsigned int i = 0; i < m_network.size(); i++)
                {
                    unsigned int nbSynapses = m_network[i].synapses.size();
                    for(unsigned int j = 0; j < nbSynapses; j++)
                    {
                        if(m_network[i].synapses[j].neurone_id > nid)
                            m_network[i].synapses[j].neurone_id--;
                        else if(m_network[i].synapses[j].neurone_id == nid)
                        {
                            for(unsigned int k = 0; k < sorties.size(); k++) // on reconnecte le neurone
                            {
                                Synapse s;
                                s.neurone_id = (sorties[k].neurone_id > nid) ? sorties[k].neurone_id - 1 : sorties[k].neurone_id;
                                s.poids = sorties[k].poids * m_network[i].synapses[j].poids;
                                m_network[i].synapses.push_back(s);
                            }
                            m_network[i].synapses.erase(m_network[i].synapses.begin() + j);
                            j--; // faut y penser
                            nbSynapses--;
                        }
                    }
                }
            }
        }
        else if(choix < 25)
        {
            // modification

            unsigned char choix2 = rand() % 4;
            nid = rand() % m_network.size();
            switch(choix2)
            {
                case 0:
                    if(m_network[nid].synapses.size() == 0)
                        continue;
                    m_network[nid].synapses[rand() % m_network[nid].synapses.size()].poids += (rand() % 2000) / 1000.f - 1.f;
                    break;
                case 1:
                    if(m_network[nid].synapses.size() == 0)
                        continue;
                    m_network[nid].synapses[rand() % m_network[nid].synapses.size()].poids *= (rand() % 2000) / 1000.f;
                    break;
                case 2:
                    m_network[nid].seuil += (rand() % 2000) / 1000.f - 1.f;
                    break;
                case 3:
                    m_network[nid].seuil *= (rand() % 2000) / 1000.f;
                    break;
            }
        }
        else
        {
            // insertion

            if(rand() % 2 == 0) // ajout d'une synapse à un neurone
            {
                nid = rand() % m_network.size();
                Synapse s;
                s.poids = (rand() % 5000) / 1000. - 2.;
                std::vector<unsigned int> neurones_suivant;
                get_neurones_after(nid, &neurones_suivant);
                if(neurones_suivant.size() < 2)
                    continue;
                s.neurone_id = neurones_suivant[rand() % (neurones_suivant.size() - 1) + 1];
                m_network[nid].synapses.push_back(s);
            }
            else // on transforme un synapse en complexe synapse-neurone-synapse
            {
                unsigned int nid = rand() % m_network.size();
                if(m_network[nid].synapses.size() < 1)
                    continue;
                unsigned int sid = rand() % m_network[nid].synapses.size();

                Neurone n;
                n.synapses = std::vector<Synapse>();
                n.type = rand() % 7; //rand() % (rand() % 9 + 1);
                n.seuil = (rand() % 2000) / 1000.f;
                Synapse s;
                s.neurone_id = m_network[nid].synapses[sid].neurone_id;
                s.poids = (rand() % 2000) / 1000.f - 1.f;
                n.synapses.push_back(s);
                m_network[nid].synapses[sid].neurone_id = m_network.size();
                m_network.push_back(n);
            }
        }

        nb_mutations++;
    }
}

void Network::get_neurones_after(unsigned int nid, std::vector<unsigned int>* all)
{
    for(unsigned int i = 0; i < all->size(); i++)
        if(all->at(i) == nid)
            return;
    all->push_back(nid);
    for(unsigned int i = 0; i < m_network[nid].synapses.size(); i++)
        get_neurones_after(m_network[nid].synapses[i].neurone_id, all);
}

QImage Network::image(qreal width, qreal height, bool progress, QWidget* parent)
{
    QProgressDialog* progressDlg;
    if(progress)
        progressDlg = new QProgressDialog("Image en cours de génération...", "Annuler", 0, height, parent);
    QImage image = QImage(width, height, QImage::Format_RGB32);
    QQueue<qint16> next;
    Neurone* neurone;
    qreal valeur;
    for(unsigned int y = 0; y < height; y++)
    {
        if(progress)
        {
            progressDlg->setValue(y);
            if(progressDlg->wasCanceled())
            {
                delete progressDlg;
                return QImage();
            }
            QApplication::processEvents();
        }
        for(unsigned int x = 0; x < width; x++)
        {
            // CALCUL DE LA COULEUR DU PIXAÏLE
            next.clear();
            next.enqueue(0);
            next.enqueue(1);
            m_network[0].valeur = x / width;
            m_network[1].valeur = y / height;
            for(unsigned int i = 2; i < m_network.count(); i++)
                m_network[i].valeur = 0;

            // initialisé, let's calcul
            while(! next.empty())
            {
                neurone = &m_network[next.dequeue()];
                switch(neurone->type)
                {
                    case 0: valeur = neurone->valeur * neurone->seuil; break;
                    case 1: valeur = (neurone->valeur < neurone->seuil) ? 0.f : 1.f; break;
                    case 2: valeur = (neurone->valeur > 0) ? pow(neurone->valeur, neurone->seuil) : pow(neurone->valeur, int(neurone->seuil)); break;
                    case 3: valeur = fmod(neurone->valeur, neurone->seuil); break;
                    case 4: valeur = SAFE_DIV(neurone->seuil, neurone->valeur); break;
                    case 5: valeur = cos(neurone->valeur); break;
                    case 6: valeur = exp(neurone->valeur); break;
                    case 7: valeur = (neurone->valeur == 0) ? 0 :log(abs(neurone->valeur)) * (float(neurone->valeur > 0)*2.f - 1.f); break;
                    case 8: valeur = (neurone->valeur > -1 and neurone->valeur < 1) ? acos(neurone->valeur) : 0; break;
                    default: std::cout << "Type de neurone invalide." << std::endl; valeur = 0; break;
                }

                neurone->valeur = 0;
                if(valeur == 0)
                    continue;

                for(qint16 i = 0; i < neurone->synapses.size(); i++)
                {
                    unsigned int neurone_id = neurone->synapses[i].neurone_id;
                    m_network[neurone_id].valeur += neurone->synapses[i].poids * valeur;
                    if(neurone->synapses[i].neurone_id > 4)
                    {
                        for(unsigned int j = 0; j < next.size(); j++)
                            if(next[j] == neurone_id)
                                next.erase(next.begin()+j);
                        next.push_back(neurone_id);
                    }
                }
            }
            // Voilà, on connait la couleur du pixel :)
            for(unsigned int i = 2; i < 5; i++)
            {
                m_network[i].valeur *= 255.;
                if(m_network[i].valeur > 255)
                    m_network[i].valeur = 255;
                if(m_network[i].valeur < 0)
                    m_network[i].valeur = 0;
            }
            image.setPixel(x, y, QColor(m_network[2].valeur, m_network[3].valeur, m_network[4].valeur).rgb());
        }
    }
    if(progress)
        delete progressDlg;
    return image;
}

void Network::empty()
{
    m_network.clear();
    Neurone neurone;
    neurone.type = 0;
    neurone.seuil = 0.5;
    neurone.synapses = std::vector<Synapse>();
    Synapse s;
    s.poids = 1;
    s.neurone_id = 2;
    neurone.synapses.push_back(s);
    s.neurone_id = 3;
    neurone.synapses.push_back(s);
    s.neurone_id = 4;
    neurone.synapses.push_back(s);
    m_network.append(neurone);
    m_network.append(neurone);
    neurone.synapses.clear();
    m_network.append(neurone);
    m_network.append(neurone);
    m_network.append(neurone);
}

bool Network::isValidNet(QList<Neurone>* network)
{
    QQueue<qint16> next;
    QList<qint16>  all;
    next.enqueue(0);
    next.enqueue(1);
    const Neurone* neurone;
    qint16 nid;
    while(! next.empty())
    {
        nid = next.dequeue();
        neurone = &network->at(nid);
        all.append(nid);
        if(all.count(nid) > 10)
            return false;
        for(unsigned int i = 0; i < neurone->synapses.size(); i++)
        {
            if(neurone->synapses[i].neurone_id > 4 && ! next.contains(neurone->synapses[i].neurone_id))
                next.enqueue(neurone->synapses[i].neurone_id);
        }
    }
    return true;
}
